import keyboard as kb
from capylayer.models.models import KeyLayer, SWITCH_MODE_NAME, LOCK_MODE_NAME

def set_key_layer_state(key_layer: KeyLayer, activate: bool) -> None:
    """   
    Map/unmap individual keys that form a key layer.
    """
    key_layer.is_active = activate

    for key_src, key_dst in key_layer.key_remaps.items():
        if activate:
            kb.remap_key(key_src, key_dst)
        else:
            kb.unremap_key(key_src) # type: ignore

def handle_modifier_lock(key_layer: KeyLayer) -> None:
    """   
    Activates/deactivates key layer based on modifier hotkey lock mode
    """
    if all(key_layer.mod_hotkey_dict.values()):
        set_key_layer_state(key_layer, not key_layer.is_active)
        for key in key_layer.mod_hotkey_dict.keys():
            key_layer.mod_hotkey_dict[key] = False

def handle_modifier_switch(key_layer: KeyLayer) -> None: 
    """   
    Activates/deactivates key layer based on modifier hotkey switch mode.
    """
    if all(key_layer.mod_hotkey_dict.values()):
        if not key_layer.is_active:
            set_key_layer_state(key_layer, True)
    else:
        if key_layer.is_active:
            set_key_layer_state(key_layer, False)

def handle_mod_mode(key_layer: KeyLayer) -> None:
    """   
    Calls corresponding handler modifier function based on modifier hotkey mode.
    """    
    if key_layer.mod_mode == SWITCH_MODE_NAME:
        handle_modifier_switch(key_layer)   
    elif key_layer.mod_mode == LOCK_MODE_NAME:
        handle_modifier_lock(key_layer)

def handle_mod_hotkey(event: kb.KeyboardEvent, key_layers: list[KeyLayer]) -> None:
    """   
    Handle key events to track press and release of keys that make up the modifier hotkey.
    """
    
    # bool on key_layer.mode_hotkey_dict is used differently depending on key_layer.mod_mode
    for key_layer in key_layers:
        if event.scan_code in key_layer.mod_hotkey_dict:
            key_layer.mod_hotkey_dict[event.scan_code] = (event.event_type == kb.KEY_DOWN)
            handle_mod_mode(key_layer)    


from typing import TypeAlias, Self, Literal
from pydantic import BaseModel, ConfigDict, model_validator, Field, conlist
import keyboard as kb

# Type aliases
ModHotkeyValues: TypeAlias = list[str]
KeyLayersValues: TypeAlias = list[dict[str, ModHotkeyValues | str | dict[str, str]]]
ProfilesValues: TypeAlias = dict[str, KeyLayersValues]
CommandsJson: TypeAlias = dict[str, dict[str, str]]

# Constants
SWITCH_MODE_NAME = "switch"
LOCK_MODE_NAME = "lock"
INDENT_STR = "   "

class ConfigModel(BaseModel):
    model_config = ConfigDict(extra = "forbid", strict = True, revalidate_instances="always")

class KeyLayer(ConfigModel):
    mod_hotkey: conlist(str, min_length = 1) # type: ignore
    mod_hotkey_dict: dict[int, bool] = Field(default = {}, repr = False, exclude = True)
    mod_mode: Literal[SWITCH_MODE_NAME, LOCK_MODE_NAME] # type: ignore
    key_remaps: dict[str, str]
    is_active: bool = Field(default = False, repr = False, exclude = True)
        
    @model_validator(mode = "after")
    def build_mod_hotkey_dict(self) -> Self:
        """
        Builds a dictionary for easier tracking of key presses of keys 
        contained in the modifier hotkey.
        """

        # keyboard.key_to_scan_codes() returns a tuple, where first item is the key scan code
        self.mod_hotkey_dict = {kb.key_to_scan_codes(key)[0]: False for key in self.mod_hotkey}
        return self
    
    def __setattr__(self, name, value):
        """
        Calls build_mod_hotkey_dict() if the attribute being set is mod_hotkey.
        """
        super().__setattr__(name, value)
        if name == "mod_hotkey":
            self.build_mod_hotkey_dict()

    def __str__(self, indent_quant: int = 0):
        indent = indent_quant * INDENT_STR
        sub_indent = (indent_quant + 1) * INDENT_STR

        key_layer_str = f"{indent}├──mod_hotkey: {self.mod_hotkey}"
        key_layer_str += f"\n{indent}├──mod_mode: {self.mod_mode}"
        key_layer_str += f"\n{indent}└──key_remaps:"
        key_layer_str += ''.join(f"\n{sub_indent}{key_src}: {key_dst}" for key_src, key_dst in self.key_remaps.items())
    
        return key_layer_str


class Profile(ConfigModel):
    name: str
    key_layers: list[KeyLayer]

    def __str__(self, indent_quant: int = 0):
        indent = indent_quant * INDENT_STR

        profile_str = f"{indent}├──name: {self.name}"
        profile_str += f"\n{indent}└──key_layers:"
        profile_str += ''.join(f"\n{key_layer.__str__(indent_quant + 1)}" for key_layer in self.key_layers)

        return profile_str

class Profiles(ConfigModel):
    active_profile_name: str
    profiles: dict[str, Profile]

    def __str__(self, indent_quant: int = 0):
        indent = indent_quant * INDENT_STR

        profiles_str = f"{indent}├──active_profile_name: {self.active_profile_name}"
        profiles_str += f"\n{indent}└──profiles:" + ''.join(f"\n{profile.__str__(indent_quant + 1)}" for profile in self.profiles.values())

        return profiles_str

class CommandsItem(ConfigModel):
    hotkey: conlist(str, min_length = 1)  # type: ignore
    hotkey_str: str = Field(default = "", repr = False, exclude = True)

    @model_validator(mode = "after")
    def build_command_hotkey_str(self) -> Self:
        """
        Transforms a list[str] to a str joined by "+" 
         (keyboard library's format for hotkeys) 
        """
        self.hotkey_str = kb.get_hotkey_name(self.hotkey)
        return self

    def __setattr__(self, name, value):
        """
        Calls build_mod_hotkey_dict() if the attribute being set is mod_hotkey.
        """
        super().__setattr__(name, value)
        if name == "hotkey":
            self.build_command_hotkey_str()

    def __str__(self, indent_quant: int = 0):
        indent = indent_quant * INDENT_STR

        commands_str = f"{indent}└──hotkey: {self.hotkey}"

        return commands_str

class Commands(ConfigModel):
    quit: CommandsItem

    def __str__(self, indent_quant: int = 0):
        indent = indent_quant * INDENT_STR

        commands_str = f"{indent}└──quit:\n{self.quit.__str__(indent_quant + 1)}"

        return commands_str



import keyboard as kb
from capylayer.controllers.key_handler import handle_mod_hotkey
from capylayer.controllers.models_handler import read_active_profile, read_commands


def main() -> None:
    profile = read_active_profile()
    commands = read_commands()

    print(f"Loaded profile:\n{profile}")
    print(f"Loaded commands:\n{commands}")

    if profile:
        kb.hook(lambda event:handle_mod_hotkey(event, profile.key_layers))

    if commands:
        print(f"\nPress \"{commands.quit.hotkey_str}\" to quit")
        kb.wait(commands.quit.hotkey_str)

if __name__ == "__main__":
    main()


from typing import Any, Type, TypeVar, cast
from pydantic import ValidationError
from capylayer.models.models import Profiles, Commands
from pathlib import Path

# Type Aliases
ModelType = Type[Profiles] | Type[Commands]
T = TypeVar('T', bound = Profiles | Commands)

# Constants
DEFAULT_QUIT_HOTKEY: list[str] = ["ctrl", "shift", "caps lock"]

def read_config_file(file_path: Path, model_type: ModelType) -> T | None:
    """   
    Reads a config file and returns it as a model.
    """ 
    try:
        with open(file_path, "r") as file:
            return cast(T, model_type.model_validate_json(file.read()))
        
    except FileNotFoundError:
        print(f"Error: File not found in {file_path}")
        return None
    except (ValidationError, Exception) as err:
        print(f"Error: {err}")
        return None
    
def write_config_file(file_path: Path, model_type: ModelType, data: dict) -> bool | None:
    """   
    Validates data against model and writes to file.
    """ 
    try:
        with open(file_path, 'w') as file:
            data_model = model_type.model_validate(data)
            file.write(data_model.model_dump_json(indent = 4))
            return True

    except FileNotFoundError:
        print(f"Error: File not found in {file_path}")
        return None
    except (ValidationError, Exception) as err:
        print(f"Error: {err}")
        return None

def edit_config_key(file_path: Path, model_type: ModelType, nested_keys: list[str], value: Any) -> bool | None:
    """   
    Writes data (value) to a nested key to a config file.
    """
    model = read_config_file(file_path, model_type)
    if not model:
        return None
    
    data_dict = model.model_dump() 

    target_dict = data_dict
    try:
        for key in nested_keys[:-1]:
            if not key in target_dict or not isinstance(target_dict[key], dict):
                raise KeyError(f"Key \"{key}\" does not exist in {file_path}")
            
            target_dict = target_dict[key]

        target_dict[nested_keys[-1]] = value

        return write_config_file(file_path, model_type, data_dict)
    
    except KeyError as err:
        print(f"Error: {err}")
        return None
    except Exception as err:
        print(f"Error: {err}")
        return None
    
def remove_config_key(file_path: Path, model_type: ModelType, nested_keys: list[str]) -> bool | None:
    """   
    Removes a nested key from config file.
    """
    model = read_config_file(file_path, model_type)
    if not model:
        return None

    data_dict = model.model_dump() 

    target_dict = data_dict

    try:
        for key in nested_keys[:-1]:
            if not key in target_dict or not isinstance(target_dict[key], dict):
                raise KeyError(f"Key \"{key}\" does not exist in {file_path} following {nested_keys}")
            
            target_dict = target_dict[key]

        
        del target_dict[nested_keys[-1]]

        return write_config_file(file_path, model_type, data_dict)
    
    except KeyError as err:
        print(f"Error: {err}")
        return None
    except (ValidationError, Exception) as err:
        print(f"Error: {err}")
        return None


from capylayer.models.utils import edit_config_key, read_config_file
from capylayer.models.models import Profiles, Profile, Commands
from typing import cast
from pathlib import Path
from importlib.resources import files

# Config files
config_dir_traversable = files("capylayer.models.config")
config_dir = cast(Path, config_dir_traversable)
profiles_path = config_dir / "profiles.json"
commands_path = config_dir / "commands.json"

def read_active_profile(file_path: Path = profiles_path) -> Profile | None:
    """   
    Returns a Profile model of the current active profile from file.
    """
    profiles = read_config_file(file_path, Profiles)
    if not profiles:
        return None

    try:
        active_profile_name = profiles.active_profile_name
        if not any(active_profile_name == profile_names for profile_names in profiles.profiles.keys()):
            print(f"No active profile called \"{active_profile_name}\" found in {file_path}.")
            print("Defaulting to first profile.")
            active_profile_name = next(iter(profiles.profiles))
            if not edit_config_key(file_path, Profiles, ["active_profile_name"], active_profile_name):
                return None
        
        active_profile = profiles.profiles[active_profile_name]
        return active_profile
    
    except Exception as err:
        print(f"Error: {err}")
        return None

def save_profile(profile: Profile, file_path: Path = profiles_path) -> bool | None:
    """   
    Saves a profile to file.
        
    If a profile in the file contains the same name as the given profile, 
    it overwrites it.
    """
    return edit_config_key(file_path, Profiles, ["profiles", f"{profile.name}"], profile)

def switch_profile(profile_name: str, file_path: Path = profiles_path) -> Profile | None:
    """   
    Switches to profile with the given name.
    """
    if not edit_config_key(file_path, Profiles, ["active_profile_name"], profile_name):
        return None
    
    return read_active_profile()

def remove_profile(profile_name: str, file_path: Path = profiles_path) -> bool | None:
    """
    Removes a profile from file.
    """
    return edit_config_key(file_path, Profiles, ["profiles", f"{profile_name}"], "")

def read_commands(file_path: Path = commands_path) -> Commands | None:
    """   
    Returns a Commands model from file.
    """
    try:
        return read_config_file(file_path, Commands)
    
    except Exception as err:
        print(f"Error: {err}")
        return None
    
def save_commands(commands: Commands, file_path: Path = commands_path) -> bool | None:
    """   
    Saves quit hotkey
    """
    return edit_config_key(file_path, Commands, ["quit", "hotkey"], commands.quit.hotkey)